@@ -1,4 +1,5 @@
 use std::io::{self, Read};
+use std::sync::Arc;
 
 use ll;
 use ::parse_code;
@@ -34,42 +35,66 @@ pub struct Decoder<R: Read> {
     offset: usize,
     // decompression context
     context: DecoderContext,
+
+    dictionary: Arc<Vec<u8>>,
+    // `true` if we should stop after the first frame.
+    single_frame: bool,
 }
 
 impl<R: Read> Decoder<R> {
     /// Creates a new decoder.
     pub fn new(reader: R) -> io::Result<Self> {
-        let context = DecoderContext::default();
+        Self::with_dictionary(reader, Vec::new())
+    }
 
-        try!(parse_code(unsafe { ll::ZSTD_initDStream(context.s) }));
-
-        Decoder::with_context(reader, context)
+    /// Sets this `Decoder` to stop after the first frame.
+    ///
+    /// By default, it keeps concatenating frames until EOF is reached.
+    pub fn single_frame(mut self) -> Self {
+        self.single_frame = true;
+        self
     }
 
     /// Creates a new decoder, using an existing dictionary.
     ///
     /// The dictionary must be the same as the one used during compression.
-    pub fn with_dictionary(reader: R, dictionary: &[u8]) -> io::Result<Self> {
-        let context = DecoderContext::default();
-
-        try!(parse_code(unsafe {
-            ll::ZSTD_initDStream_usingDict(context.s,
-                                           dictionary.as_ptr(),
-                                           dictionary.len())
-        }));
-
-        Decoder::with_context(reader, context)
+    pub fn with_dictionary<V>(reader: R, dictionary: V) -> io::Result<Self>
+        where V: Into<Vec<u8>>
+    {
+        Self::with_shared_dictionary(reader, Arc::new(dictionary.into()))
     }
 
-    fn with_context(reader: R, context: DecoderContext) -> io::Result<Self> {
+    /// Creates a new decoder, using a shared dictionary.
+    ///
+    /// The dictionary must be the same as the one used during compression.
+    ///
+    /// The dictionary can be shared with other `Decoder`s / `Encoder`s / ...
+    pub fn with_shared_dictionary(reader: R, dictionary: Arc<Vec<u8>>)
+                                  -> io::Result<Self> {
+
         let buffer_size = unsafe { ll::ZSTD_DStreamInSize() };
 
-        Ok(Decoder {
+        let mut decoder = Decoder {
             reader: reader,
             buffer: Vec::with_capacity(buffer_size),
+            dictionary: dictionary,
             offset: 0,
-            context: context,
-        })
+            context: DecoderContext::default(),
+            single_frame: false,
+        };
+
+        try!(decoder.init());
+
+        Ok(decoder)
+    }
+
+    fn init(&mut self) -> io::Result<()> {
+        try!(parse_code(unsafe {
+            ll::ZSTD_initDStream_usingDict(self.context.s,
+                                           self.dictionary.as_ptr(),
+                                           self.dictionary.len())
+        }));
+        Ok(())
     }
 
     /// Recommendation for the size of the output buffer.
@@ -84,11 +109,39 @@ impl<R: Read> Decoder<R> {
     pub fn finish(self) -> R {
         self.reader
     }
+
+    fn refill_buffer(&mut self, in_buffer: &mut ll::ZSTD_inBuffer)
+                     -> io::Result<bool> {
+
+        // We need moar data!
+        // Make a nice clean buffer
+        let buffer_size = self.buffer.capacity();
+        unsafe {
+            self.buffer.set_len(buffer_size);
+        }
+
+        // And FILL IT!
+        let read = try!(self.reader.read(&mut self.buffer));
+        unsafe {
+            self.buffer.set_len(read);
+        }
+        in_buffer.pos = 0;
+        in_buffer.size = read;
+        // So we can't read anything: input is exhausted.
+
+        Ok(read > 0)
+    }
 }
 
 impl<R: Read> Read for Decoder<R> {
     fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
 
+        let mut in_buffer = ll::ZSTD_inBuffer {
+            src: self.buffer.as_ptr(),
+            size: self.buffer.len(),
+            pos: self.offset,
+        };
+
         if self.offset > self.buffer.capacity() {
             // If we've reached the end of the frame before,
             // don't even try to read more.
@@ -100,34 +153,12 @@ impl<R: Read> Read for Decoder<R> {
             size: buf.len(),
             pos: 0,
         };
-        let mut in_buffer = ll::ZSTD_inBuffer {
-            src: self.buffer.as_ptr(),
-            size: self.buffer.len(),
-            pos: self.offset,
-        };
         while out_buffer.pos != buf.len() {
 
             let mut input_exhausted = false;
 
             if in_buffer.pos == in_buffer.size {
-                // We need moar data!
-                // Make a nice clean buffer
-                let buffer_size = self.buffer.capacity();
-                unsafe {
-                    self.buffer.set_len(buffer_size);
-                }
-
-                // And FILL IT!
-                let read = try!(self.reader.read(&mut self.buffer));
-                unsafe {
-                    self.buffer.set_len(read);
-                }
-                in_buffer.pos = 0;
-                in_buffer.size = read;
-                // So we can't read anything: input is exhausted.
-                if read == 0 {
-                    input_exhausted = true;
-                }
+                input_exhausted = !try!(self.refill_buffer(&mut in_buffer));
             }
 
             let res = unsafe {
@@ -147,8 +178,20 @@ impl<R: Read> Read for Decoder<R> {
             if res == 0 {
                 // Remember that we've reached the end of the current frame,
                 // so we don't try to read the next one.
-                in_buffer.pos = self.buffer.capacity() + 1;
-                break;
+                if self.single_frame {
+                    in_buffer.pos = self.buffer.capacity() + 1;
+                    break;
+                } else {
+                    if in_buffer.pos == in_buffer.size &&
+                       !try!(self.refill_buffer(&mut in_buffer)) {
+                        // we're out.
+                        in_buffer.pos = self.buffer.capacity() + 1;
+                        break;
+                    } else {
+                        // ?
+                        try!(self.init());
+                    }
+                }
             }
         }
         self.offset = in_buffer.pos;
